// Jrn Loviscach, 2001-06-10

package doosabin;

import java.awt.*;
import javax.swing.*;
import java.awt.event.*;

public class DooSabin {
  SubdivObject[] myObject;

  public static void main(String[] args) {
    new DooSabin();
  }

  public DooSabin() {

    myObject = new SubdivObject[4];
    myObject[0] = SimpleObject.BuildCube(1.0);
    myObject[1] = myObject[0].Subdivide();
    myObject[2] = myObject[1].Subdivide();
    myObject[3] = myObject[2].Subdivide();

    JFrame myFrame = new JFrame("Doo-Sabin Subdivision");
    myFrame.addWindowListener(new WindowAdapter(){public void windowClosing(WindowEvent e){System.exit(0);}});

    SubdivPanel upperLeft = new SubdivPanel(0, myObject, myFrame);
    SubdivPanel upperRight = new SubdivPanel(1, myObject, myFrame);
    SubdivPanel lowerLeft = new SubdivPanel(2, myObject, myFrame);
    SubdivPanel lowerRight = new SubdivPanel(3, myObject, myFrame);

    JSplitPane upperSplit = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, upperLeft, upperRight);
    upperSplit.setDividerSize(3);
    JSplitPane lowerSplit = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, lowerLeft, lowerRight);
    lowerSplit.setDividerSize(3);
    JSplitPane upperLowerSplit = new JSplitPane(JSplitPane.VERTICAL_SPLIT, upperSplit, lowerSplit);
    upperLowerSplit.setDividerSize(3);
    myFrame.getContentPane().add(upperLowerSplit);

    myFrame.setSize(500, 500);
    myFrame.setLocation(100, 100);
    myFrame.show();

    // Schieberegler lsst sich dummerweise erst positionieren, wenn angezeigt
    upperSplit.setDividerLocation(0.5);
    lowerSplit.setDividerLocation(0.5);
    upperLowerSplit.setDividerLocation(0.5);

    JOptionPane.showMessageDialog(myFrame,
        "This software demonstrates subdivision surfaces\n" +
        "generated by the Doo-Sabin algorithm.\n\n" +
        "How to use: Drag vertices in upper left view.\n" +
        "Drag Split lines to change view sizes.\n\n" +
        "Jrn Loviscach, 2001-06-10",
        "How to Use", JOptionPane.INFORMATION_MESSAGE);
  }
}

class SubdivPanel
extends JPanel
{
  int level;
  SubdivObject[] myObject;
  JFrame myFrame;

  SubdivPanel(int l, SubdivObject[] o, JFrame f)
  {
    level = l;
    myObject = o;
    myFrame = f;

    // fr erstes Panel MouseMotionListener anlegen
    if( l == 0 )
    {
        addMouseMotionListener (new VertexListener(myObject[0].vertices, myObject, this, myFrame) );
    }
  }

  public void paint(Graphics g)
  {
    super.paint(g);
    // Komponenten zeichen (hier nur: grauer Hintergrund)

    int w = getWidth();
    int h = getHeight();
    g.clearRect(0, 0, w, h); // zu Fu lschen

    Edge[] e = myObject[level].edges;
    Vertex[] v = myObject[level].vertices;

    Graphics2D g2 = (Graphics2D) g;
    // Graphics2D-Grafikkontext fr besondere Java2D-Grafikfunktionen
    // hier: Antialiasing

    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    //Kantenglttung anschalten

    for( int i=0; i < e.length; ++i)
    {
      Vertex v0 = v[e[i].vertex0];
      Vertex v1 = v[e[i].vertex1];

      double zMean = (v0.z + v1.z)/2;
      int c = (int)(120.0 + 50.0*zMean);
      g2.setColor(new Color(c, c, c));
      g2.setStroke(new BasicStroke((float)(2.0-1.4*zMean)));
      // der BasicStroke-Konstruktor will float statt double

      Point point1 = v0.getPoint(w, h);
      Point point2 = v1.getPoint(w, h);
      g2.drawLine(point1.x, point1.y, point2.x, point2.y);
    }

    for( int i=0; i < v.length; ++i)
    {
      int r = (int)(128 - 200*v[i].z);
      int b = (int)(128 + 128*v[i].z);
      g2.setPaint(new Color(r, 128, b));

      Point myPoint = v[i].getPoint(w, h);

      if(level == 0) // Punkte oben links mit groen Quadraten markieren, sonst Kreise
      {
        g2.fillRect(myPoint.x-4, myPoint.y-4, 9, 9);
      }
      else
      {
        int s = (int)(2.0-0.8*v[i].z);
        g2.fillOval(myPoint.x-s, myPoint.y-s, 2*s+1, 2*s+1);
      }
    }
  }
}

class VertexListener
implements MouseMotionListener
{
  Vertex[] myVertices;
  SubdivObject[] myObject;
  SubdivPanel myPanel;
  Frame myFrame;
  int xLast = -1;
  int yLast = -1;
  int draggedVertex = -1;

  VertexListener( Vertex[] v, SubdivObject[] o, SubdivPanel p, JFrame f )
  {
    myVertices = v;
    myObject = o;
    myPanel = p;
    myFrame = f;
  }

  public void mouseMoved(MouseEvent e)
  {
    xLast = e.getX();
    yLast = e.getY();

    for( int k=0; k<myVertices.length; ++k)
    {
      Point p = myVertices[k].getPoint(myPanel.getWidth(), myPanel.getHeight());
      if( Math.abs(p.x - xLast)<5 && Math.abs(p.y - yLast)<5 )
      {
        draggedVertex = k;
        break;
      }
    }
  }

  public void mouseDragged(MouseEvent e)
  {
    int x = e.getX();
    int y = e.getY();
    myVertices[draggedVertex].move(x - xLast, y - yLast, myPanel.getWidth(), myPanel.getHeight());
    xLast = x;
    yLast = y;

    // Krperkoordinaten neu berechnen
    myObject[1] = myObject[0].Subdivide(myObject[1]);
    myObject[2] = myObject[1].Subdivide(myObject[2]);
    myObject[3] = myObject[2].Subdivide(myObject[3]);

    myFrame.repaint(); // neuzeichnen
  }
}

class Vertex // Eckpunkt
{
  double x, y, z;

  Vertex(double xi, double yi, double zi)
  {
    x = xi;
    y = yi;
    z = zi;
  }

  Point getPoint(int w, int h) // hole projizierte Koordinaten, skaliert auf Fenster
  {
    int d = (w<h) ? w : h; // die kleinere der beiden Fensterkanten
    return new Point( w/2 + (int)(d*(0.5*x+0.08*z)/(1.0+z/8.0)), h/2 + (int)(d*(-0.5*y-0.08*z)/(1.0+z/8.0)) );
  }

  void move(int dx, int dy, int w, int h)
  {
    int d = (w<h) ? w : h; // die kleinere der beiden Fensterkanten
    // von nderung der Bildschirmkoordinaten auf nderung der
    // 3D-Koordinaten zurckrechnen
    x += dx/0.5/d*(1.0+z/8.0);
    y -= dy/0.5/d*(1.0+z/8.0);
  }
}

class Edge // Kante
{
  int vertex0, vertex1;

  Edge(int v1, int v2)
  {
    vertex0 = v1;
    vertex1 = v2;
  }
}

class Face // Polygon
{
  int[] edges;

  Face(int[] e)
  {
    edges = e;
  }
}

class SubdivObject
{
  // Boundary Representation:
  Vertex[] vertices;
  Edge[] edges;
  Face[] faces;

  SubdivObject(Vertex[] v, Edge[] e, Face[] f)
  {
    vertices = v;
    edges = e;
    faces = f;
  }

  SubdivObject Subdivide()
  {
    // Arraygren fr unterteiltes Objekt bestimmen
    int numberOfNewVertices = 0;
    for( int i=0; i < faces.length; ++i ) // alle Polygone
    {
      numberOfNewVertices += faces[i].edges.length;
    }
    Vertex[] newVertices = new Vertex[numberOfNewVertices];
    Edge[] newEdges = new Edge[2*numberOfNewVertices];
    Face[] newFaces = new Face[faces.length + edges.length + vertices.length];

    int[][] newVertexFromOldFace = new int[faces.length][];
    // newVertexFromOldFace[i][j] liefert nach dem Berechnen der neuen Punkte
    // die Nummer des neuen Punkts in newVertices[], der aus dem alten Polygon i
    // und dessen Punkt j entstanden ist
    // (j = 0, 1, ... Zahl der Punkte dieses Polygons minus 1.)

    int[][] oldVertexFromOldFace = new int[faces.length][];
    // hnlich, nur dass jeweils die alten Eckpunkte herauskommen

    // Koordinaten der neuen Punkte berechnen
    int countConstructedVertices = 0;
    for( int i=0; i < faces.length; ++i ) // alle Polygone
    {
      int ngon = faces[i].edges.length;
      // Zahl der Eckpunkte bzw. Kanten des aktuellen Polygons

      newVertexFromOldFace[i] = new int[ngon];
      oldVertexFromOldFace[i] = new int[ngon];

      // suche alte Punkte fr dieses Polygon zusammen
      for ( int j=0; j < ngon; j++ )
      {
        // finde den gemeinsamen Punkt von Kante j und ihrem Nachfolger
        // (Kanten sind ja im Polygon zyklisch aufgelistet)
        Edge thisEdge = edges[ faces[i].edges[  j         ] ];
        Edge nextEdge = edges[ faces[i].edges[ (j+1)%ngon ] ];
        // modulo, damit Nachfolger von ngon-1 die Nummer 0 ist

        int commonVertex;
        if (thisEdge.vertex0 == nextEdge.vertex0 || thisEdge.vertex0 == nextEdge.vertex1)
        {
          commonVertex = thisEdge.vertex0;
        }
        else
        {
          commonVertex = thisEdge.vertex1;
        }

        oldVertexFromOldFace[i][j] = commonVertex;
      }

      // bestimme Mittelwert aller Eckpunkte des Polygons
      double xs = 0.0, ys = 0.0, zs = 0.0;
      for ( int j=0; j < ngon; j++ )
      {
        xs += vertices[oldVertexFromOldFace[i][j]].x;
        ys += vertices[oldVertexFromOldFace[i][j]].y;
        zs += vertices[oldVertexFromOldFace[i][j]].z;
      }
      double xm = xs / ngon;
      double ym = ys / ngon;
      double zm = zs / ngon;

      // gehe Eckpunkte durch und bestimme jeweils zugehrigen neuen Punkt
      for( int j=0; j < ngon; j++)
      {
        // Summiere Mittelpunkt und aktuellen Punkt.
        // 2.0 beim aktuellen Punkt und 0.5 bei den beiden Nachbarn (folgende Schleife)
        // entspricht 1.0 mal aktuellen Punkt plus die Mitten der benachbarten Kanten.
        double xsum = xm + 2.0*vertices[oldVertexFromOldFace[i][j]].x;
        double ysum = ym + 2.0*vertices[oldVertexFromOldFace[i][j]].y;
        double zsum = zm + 2.0*vertices[oldVertexFromOldFace[i][j]].z;

        // summiere die jeweils anderen Punkte der beiden angrenzenden Kanten des Polygons
        int precedingVertex = oldVertexFromOldFace[i][ (j==0) ? (ngon-1) : (j-1) ];
        // Vorgnger von Punkt 0 ist der letzte Punkt, also Nummer oldFaceVertices.length-1
        xsum += 0.5*vertices[precedingVertex].x;
        ysum += 0.5*vertices[precedingVertex].y;
        zsum += 0.5*vertices[precedingVertex].z;
        int followingVertex = oldVertexFromOldFace[i][ (j+1) % ngon ];
        // Nachfolger des letzten Punkts ist Punkt 0
        xsum += 0.5*vertices[followingVertex].x;
        ysum += 0.5*vertices[followingVertex].y;
        zsum += 0.5*vertices[followingVertex].z;

        newVertices[countConstructedVertices] = new Vertex(xsum/4, ysum/4, zsum/4);
        newVertexFromOldFace[i][j] = countConstructedVertices;
        ++countConstructedVertices;
      }
    }

    // Wir erhalten in diesem Schritt zwei alte Kanten unter jeder alten,
    // eine unter jedem benachbarten Polygon.
    // Diese Verwandtschaft merken wir uns:
    int[][] newEdgeFromOldEdge = new int[edges.length][2];
    // Hier mit gleicher Indexierung die Nummer des alten Polygons,
    // zu dem der neue Eckpunkt gehrt:
    int[][] faceOfNewEdgeFromOldEdge = new int[edges.length][2];
    int[] newEdgeFromOldEdge_Count = new int[edges.length]; // nachher berall gleich 2

    // erzeuge fr jedes alte Polygon die neuen Kanten und das neue Polygon im Inneren
    int countConstructedEdges = 0;
    int countConstructedFaces = 0;
    for( int i=0; i < faces.length; ++i ) // alle Polygone
    {
      int[] newFaceEdges = new int[faces[i].edges.length];
      for( int j=0; j < faces[i].edges.length; ++j) // die jeweiligen Kanten
      {
        newEdges[countConstructedEdges] = new Edge(newVertexFromOldFace[i][ (j>0) ? (j-1) : (faces[i].edges.length-1) ], newVertexFromOldFace[i][ j ]);
        int edge = faces[i].edges[j];
        newEdgeFromOldEdge[edge][newEdgeFromOldEdge_Count[edge]] = countConstructedEdges;
        faceOfNewEdgeFromOldEdge[edge][newEdgeFromOldEdge_Count[edge]] = i;
        ++newEdgeFromOldEdge_Count[edge];
        newFaceEdges[j] = countConstructedEdges;
        ++countConstructedEdges;
      }
      newFaces[countConstructedFaces] = new Face(newFaceEdges);
      ++countConstructedFaces;
    }

    // Quer zu jeder alten Kanten erhalten wir in diesem Schritt zwei neue Kanten.
    // Die mssen wir uns fr den letzten Schritt merken, um dann rund um jeden
    // alten Eckpunkt herum neue Kanten anzulegen.
    // die beiden neue Quer-Kanten merken:
    int[] newEdgesFromOldEdge = new int[2*edges.length];
    // mit gleicher Indexierung die beiden alten Eckpunkte merken:
    int[] oldVertexFromNewEdge = new int[2*edges.length];

    // erzeuge fr jede alte Kante je zwei neue Kanten und das Polygon unter der Kante
    for( int j=0; j < edges.length; j++ ) // alle alten Kanten
    {
      int[] newFaceEdges = new int[4]; // neues Polygon unter der Kante ist immer Viereck

      // die beiden alten Eckpunkte
      int oldVertex0 = edges[j].vertex0;
      int oldVertex1 = edges[j].vertex1;

      // suche die beiden benachbarten alten Polygone
      int[] oldFaces = new int[2];
      int oldFaces_Count = 0; // nachher gleich 2
      for( int i=0; i<faces.length; ++i)
      {
        for( int ji=0; ji<faces[i].edges.length; ++ji)
        {
          if(j == faces[i].edges[ji])
          {
            oldFaces[oldFaces_Count++] = i;
            break; // for( ji ) beenden
          }
        }
      }

      // zwei Kanten haben wir schon,
      // edge0 soll unter altem Polygon oldFaces[0] liegen,
      // edge1 unter altem Polygon oldFaces[1]
      int edge0 = newEdgeFromOldEdge[j][0];
      int edge1 = newEdgeFromOldEdge[j][1];
      if( faceOfNewEdgeFromOldEdge[j][0] == oldFaces[1] )
      { // austauschen
        int h = edge0;
        edge0 = edge1;
        edge1 = h;
      }

      // Indizes der beiden Kanteneckpunkte in oldFaces[0]
      int vert00 = -1;
      int vert01 = -1;
      for( int ji=0; ji<faces[oldFaces[0]].edges.length; ++ji )
      {
        if( oldVertexFromOldFace[oldFaces[0]][ji] == oldVertex0 )
        {
          vert00 = ji;
        }
        else if( oldVertexFromOldFace[oldFaces[0]][ji] == oldVertex1 )
        {
          vert01 = ji;
        }
      }
      // Indizes der beiden Kanteneckpunkte in oldFaces[1]
      int vert10 = -1;
      int vert11 = -1;
      for( int ji=0; ji<faces[oldFaces[1]].edges.length; ++ji )
      {
        if( oldVertexFromOldFace[oldFaces[1]][ji] == oldVertex0 )
        {
          vert10 = ji;
        }
        else if( oldVertexFromOldFace[oldFaces[1]][ji] == oldVertex1 )
        {
          vert11 = ji;
        }
      }

      oldVertexFromNewEdge[2*j] = oldVertex0;
      oldVertexFromNewEdge[2*j+1] = oldVertex1;
      // Sortiere wegen Umlaufsinn
      // (Vorbereitung fr Backface Culling)
      if( (vert00>vert01 && vert01!=0) || (vert00==0 && vert01!=1) ) // soll zyklisch werden
      {
        int help = vert01;
        vert01 = vert00;
        vert00 = help;

        help = vert11;
        vert11 = vert10;
        vert10 = help;
        oldVertexFromNewEdge[2*j] = oldVertex1;
        oldVertexFromNewEdge[2*j+1] = oldVertex0;
      }

      // nun konstruieren wir noch die zwei neuen Kanten quer zur alten Kante
      newEdges[countConstructedEdges] =
          new Edge( newVertexFromOldFace[oldFaces[1]][vert10], newVertexFromOldFace[oldFaces[0]][vert00] );
      newEdges[countConstructedEdges + 1] =
          new Edge( newVertexFromOldFace[oldFaces[0]][vert01], newVertexFromOldFace[oldFaces[1]][vert11] );
          // Achtung: Reihenfolge der Vertizes extra so herum (_gegen_ Umlaufsinn),
          // damit Umlauf im Uhrzeigersinn um Kanten, das hilft im nchsten Schritt

      newFaceEdges[0] = countConstructedEdges;
      newEdgesFromOldEdge[2*j] = countConstructedEdges;
      newFaceEdges[1] = edge1;
      newFaceEdges[2] = countConstructedEdges + 1;
      newEdgesFromOldEdge[2*j+1] = countConstructedEdges + 1;
      newFaceEdges[3] = edge0;
      countConstructedEdges += 2;

      newFaces[countConstructedFaces++] = new Face(newFaceEdges);
    }

    // erzeuge fr jeden alten Eckpunkt das Polygon unter der Ecke
    for( int k=0; k < vertices.length; ++k ) // alle alten Eckpunkte
    {
      // finde alle beteiligten neuen Kanten
      int vertEdgesCount = 0; // Zahl der umlaufenden Kanten
      int[] vertEdges = new int[2*edges.length]; // berdimensioniert, aber Gre schwer zu schtzen (Vector benutzen?)
      for( int j=0; j < 2*edges.length; ++j ) // alle neuen Kanten unter alten Kanten
      {
        if( k == oldVertexFromNewEdge[j] )
        {
          vertEdges[vertEdgesCount++] = j;
        }
      }

      int[] newFaceEdges = new int[vertEdgesCount];
      // stckele Kanten mit passenden Eckpunkten aneinander,
      // das geht, weil die Kanten mit passender Reihenfolge
      // von vertex0 und vertex1 erzeugt worden sind
      newFaceEdges[0] = newEdgesFromOldEdge[vertEdges[0]];
      int vert = newEdges[newEdgesFromOldEdge[vertEdges[0]]].vertex1;
      for( int j=1; j<vertEdgesCount; ++j )
      {
        for( int ji=0; ji<vertEdgesCount ; ++ji )
        {
          if( vert == newEdges[newEdgesFromOldEdge[vertEdges[ji]]].vertex0 )
          {
            vert = newEdges[newEdgesFromOldEdge[vertEdges[ji]]].vertex1;
            newFaceEdges[j] = newEdgesFromOldEdge[vertEdges[ji]];
            break; // beendet for( ji )
          }
        }
      }
      newFaces[countConstructedFaces++] = new Face(newFaceEdges);
    }
    return new SubdivObject(newVertices, newEdges, newFaces);
  }

  // Wenn wir nur Eckpunkte verschieben, aber die Verbindungen (Topologie)
  // gleich lassen, mssen wir nur die neuen Eckpunkte berechnen
  // und knnen die alten Tabellen fr Kanten und Flchen weiter verwenden.
  SubdivObject Subdivide(SubdivObject previousNewObject)
  {
    Vertex[] newVertices = new Vertex[previousNewObject.vertices.length];

    int[][] oldVertexFromOldFace = new int[faces.length][];
    // newVertexFromOldFace[i][j] liefert nach dem Berechnen der neuen Punkte
    // die Nummer des alten Punkts Nummer j im alten Polygon i
    // (j = 0, 1, ... Zahl der Punkte dieses Polygons minus 1.)

    // Koordinaten der neuen Punkte berechnen
    int countConstructedVertices = 0;
    for( int i=0; i < faces.length; ++i ) // alle Polygone
    {
      int ngon = faces[i].edges.length;
      // Zahl der Eckpunkte bzw. Kanten des aktuellen Polygons

      oldVertexFromOldFace[i] = new int[ngon];

      // suche alte Punkte fr dieses Polygon zusammen
      for ( int j=0; j < ngon; j++ )
      {
        // finde den gemeinsamen Punkt von Kante j und ihrem Nachfolger
        // (Kanten sind ja im Polygon zyklisch aufgelistet)
        Edge thisEdge = edges[ faces[i].edges[  j         ] ];
        Edge nextEdge = edges[ faces[i].edges[ (j+1)%ngon ] ];
        // modulo, damit Nachfolger von ngon-1 die Nummer 0 ist

        int commonVertex;
        if (thisEdge.vertex0 == nextEdge.vertex0 || thisEdge.vertex0 == nextEdge.vertex1)
        {
          commonVertex = thisEdge.vertex0;
        }
        else
        {
          commonVertex = thisEdge.vertex1;
        }

        oldVertexFromOldFace[i][j] = commonVertex;
      }

      // bestimme Mittelwert aller Eckpunkte des Polygons
      double xs = 0.0, ys = 0.0, zs = 0.0;
      for ( int j=0; j < ngon; j++ )
      {
        xs += vertices[oldVertexFromOldFace[i][j]].x;
        ys += vertices[oldVertexFromOldFace[i][j]].y;
        zs += vertices[oldVertexFromOldFace[i][j]].z;
      }
      double xm = xs / ngon;
      double ym = ys / ngon;
      double zm = zs / ngon;

      // gehe Eckpunkte durch und bestimme jeweils zugehrigen neuen Punkt
      for( int j=0; j < ngon; j++)
      {
        // Summiere Mittelpunkt und aktuellen Punkt.
        // 2.0 beim aktuellen Punkt und 0.5 bei den beiden Nachbarn (folgende Schleife)
        // entspricht 1.0 mal aktuellen Punkt plus die Mitten der benachbarten Kanten.
        double xsum = xm + 2.0*vertices[oldVertexFromOldFace[i][j]].x;
        double ysum = ym + 2.0*vertices[oldVertexFromOldFace[i][j]].y;
        double zsum = zm + 2.0*vertices[oldVertexFromOldFace[i][j]].z;

        // summiere die jeweils anderen Punkte der beiden angrenzenden Kanten des Polygons
        int precedingVertex = oldVertexFromOldFace[i][ (j==0) ? (ngon-1) : (j-1) ];
        // Vorgnger von Punkt 0 ist der letzte Punkt, also Nummer oldFaceVertices.length-1
        xsum += 0.5*vertices[precedingVertex].x;
        ysum += 0.5*vertices[precedingVertex].y;
        zsum += 0.5*vertices[precedingVertex].z;
        int followingVertex = oldVertexFromOldFace[i][ (j+1) % ngon ];
        // Nachfolger des letzten Punkts ist Punkt 0
        xsum += 0.5*vertices[followingVertex].x;
        ysum += 0.5*vertices[followingVertex].y;
        zsum += 0.5*vertices[followingVertex].z;

        newVertices[countConstructedVertices] = new Vertex(xsum/4, ysum/4, zsum/4);
        ++countConstructedVertices;
      }
    }

    return new SubdivObject(newVertices, previousNewObject.edges, previousNewObject.faces);
  }
}

class SimpleObject
{
  static SubdivObject BuildCube(double d) // d = Kantenlnge
  {
    double s = d/2; // Koordinaten gehen von -s bis +s

    Vertex[] vertices =
    {
      new Vertex(+s, +s, +s),
      new Vertex(+s, +s, -s),
      new Vertex(+s, -s, +s),
      new Vertex(+s, -s, -s),
      new Vertex(-s, +s, +s),
      new Vertex(-s, +s, -s),
      new Vertex(-s, -s, +s),
      new Vertex(-s, -s, -s)
    };

    Edge[] edges =
    {
      new Edge(0, 1),
      new Edge(0, 2),
      new Edge(0, 4),
      new Edge(1, 3),
      new Edge(1, 5),
      new Edge(2, 3),
      new Edge(2, 6),
      new Edge(3, 7),
      new Edge(4, 5),
      new Edge(4, 6),
      new Edge(5, 7),
      new Edge(6, 7)
    };

    Face[] faces =
    { // Vorbereitung fr Backface Culling:
      // Kanten im Uhrzeigersinn (von auen gesehen) aufgelistet
      new Face( new int[]{ 0,  1,  5,  3} ),
      new Face( new int[]{ 0,  4,  8,  2} ),
      new Face( new int[]{ 1,  2,  9,  6} ),
      new Face( new int[]{ 3,  7, 10,  4} ),
      new Face( new int[]{ 5,  6, 11,  7} ),
      new Face( new int[]{ 8, 10, 11,  9} )
    };

    return new SubdivObject(vertices, edges, faces);
  }
}
